---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--          Copyright (c) 2007, Raul Fechete, Georg Kienesberger         --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Ast2Cfg.Output; use Ast2Cfg.Output; --TODO: remove line

package body Ast2Cfg.Generic_List is

   -- Free a "Container" from memory
   procedure Free_Container is new Ada.Unchecked_Deallocation(Container, Container_Ptr);

   -- Initialize list structure information
   procedure Init (Obj: out Object) is
   begin
      Obj.Head := null;
      Obj.Tail := null;
      Obj.Curr := null;
      Obj.Count := 0;
   end Init;

   -- Initialize list structure information
   procedure Init (Ptr: out Pointer) is
   begin
      if Ptr = null then
         Ptr := new Object;
      end if;
      Init(Ptr.all);
   end Init;

   -- Append a new element to the list
   procedure Append (Obj: in out Object; I: in Item) is
      Last: Container_Ptr := Obj.Tail;
   begin
      if (Last = null) then
         Obj.Tail := new Container'(Data => I, Prev => null, Succ => null);
         Obj.Head := Obj.Tail;
      else
         Last.Succ := new Container'(Data => I, Prev => Last, Succ => null);
         Obj.Tail := Last.Succ;
      end if;
      Obj.Curr := Obj.Head;
      Obj.Count := Obj.Count + 1;
   end Append;

   -- Append a new element to the list
   procedure Append (Ptr: in out not null Pointer; I: in Item) is
   begin
      Append(Ptr.all, I);
   end Append;

   procedure Append (Obj: in out Object; O: in out Object) is
      Temp_List: Object := O;
      I: Item;
   begin
      Reset(Temp_List);
      while (Has_Next(Temp_List)) loop
         Get_Next(Temp_List, I);
         Append(Obj, I);
      end loop;
   end Append;

   procedure Append (Ptr: in out not null Pointer; P: in out not null Pointer) is
   begin
      Append(Ptr.all, P.all);
   end Append;

   -- Replace the head element of the list
   procedure Replace_Head (Obj: in out Object; I: in Item) is
   begin
      if (Obj.Head /= null) then
         Obj.Head.Data := I;
      end if;
   end Replace_Head;

   procedure Replace_Head (Ptr: in out not null Pointer; I: in Item) is
   begin
      Replace_Head(Ptr.all, I);
   end Replace_Head;

   -- Insert a new element at the beginning of the list
   procedure Insert_Head (Obj: in out Object; I: in Item) is
      Aux: Container_Ptr := Obj.Head;
   begin
      if (Aux = null) then
         Append (Obj, I);
      else
         Obj.Head := new Container'(Data => I, Prev => null, Succ => Aux);
         Aux.Prev := Obj.Head;
         Obj.Curr := Obj.Head;
         Obj.Count := Obj.Count + 1;
      end if;
   end Insert_Head;

   -- Insert a new element at the beginning of the list
   procedure Insert_Head (Ptr: in out not null Pointer; I: in Item) is
   begin
      Insert_Head(Ptr.all, I);
   end Insert_Head;

   function Get_Head (Obj: in Object) return Item is
      H: Container_Ptr := Obj.Head;
   begin
      if (H /= null) then
         return H.Data;
      else
         raise List_Exception;
      end if;
   end Get_Head;

   function Get_Head (Ptr: in not null Pointer) return Item is
   begin
      return Get_Head(Ptr.all);
   end Get_Head;

   function Get_Tail (Obj: in Object) return Item is
      H: Container_Ptr := Obj.Tail;
   begin
      if (H /= null) then
         return H.Data;
      else
         raise List_Exception;
      end if;
   end Get_Tail;

   function Get_Tail (Ptr: in not null Pointer) return Item is
   begin
      return Get_Tail(Ptr.all);
   end Get_Tail;

   -- Replace the tail element
   procedure Remove_Last (Obj: in out Object) is
      J: Item;
   begin
      Pop(Obj, J);
   end Remove_Last;

   -- Replace the tail element
   procedure Remove_Last (Ptr: in out not null Pointer) is
   begin
      Remove_Last(Ptr.all);
   end Remove_Last;

   -- removes the first element equal to I
   procedure Remove (Obj: in out Object; I: in Item) is
      Cur: Item;
      Del: Container_Ptr;
   begin
      Reset(Obj);
      while (Has_Next(Obj)) loop
         Get_Next(Obj, Cur);
         if Cur = I then
            if Obj.Curr = null then
               Obj.Curr := Obj.Tail;
            else
               Obj.Curr := Obj.Curr.Prev;
            end if;

            if Obj.Curr = Obj.Head then
               if Obj.Count = 1 then
                  Free_Container(Obj.Head);
                  Init(Obj);
               else
                  Del := Obj.Head;
                  Obj.Head := Obj.Head.Succ;
                  Obj.Head.Prev := null;
                  Obj.Curr := Obj.Head;
                  Free_Container(Del);
                  Obj.Count := Obj.Count - 1;
               end if;
            elsif Obj.Curr = Obj.Tail then
               Remove_Last(Obj);
            else
               Del := Obj.Curr;
               Obj.Curr.Succ.Prev := Obj.Curr.Prev;
               Obj.Curr.Prev.Succ := Obj.Curr.Succ;
               Obj.Curr := Obj.Curr.Succ;
               Free_Container(Del);
               Obj.Count := Obj.Count - 1;
            end if;

            return;
         elsif Obj.Curr = null then
            raise List_Exception with "Element to remove not found!";
         end if;
      end loop;
   end Remove;

   -- removes the first element equal to I
   procedure Remove (Ptr: in out not null Pointer; I: in Item) is
   begin
      Remove(Ptr.all, I);
   end Remove;

   -- Create a copy of a list
   procedure Duplicate (Src: in out Object; Dst: out Object) is
      New_Obj: Object;
      I: Item;
   begin
      Init(New_Obj);
      Reset(Src);
      while (Has_Next(Src)) loop
         Get_Next(Src,I);
         Append(New_Obj,I);
      end loop;
      Dst := New_Obj;
   end Duplicate;

   -- Create a copy of a list
   procedure Duplicate (Src: in out not null Pointer; Dst: out Pointer) is
   begin
      if Dst = null then
         Dst := new Object;
         Init(Dst);
      end if;
      Duplicate(Src.all, Dst.all);
   end Duplicate;

   -- Destroy the elements of the list
   procedure Clear (Obj: in out Object) is
      Ptr,Aux: Container_Ptr;
   begin
      Ptr := Obj.Head;
      while Ptr /= null loop
         Aux := Ptr;
         Ptr := Ptr.Succ;
         Free_Container(Aux);
      end loop;
      Init(Obj);
   end Clear;

   -- Destroy the elements of the list
   procedure Clear (Ptr: in out not null Pointer) is
   begin
      Clear(Ptr.all);
   end Clear;

   procedure Free (Ptr: in out Pointer) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Object, Pointer);
   begin
      if Ptr = null then
         return;
      end if;

      Clear(Ptr.all);
      Destroy(Ptr);
   end Free;

   -- Reset the iterator
   procedure Reset (Obj: in out Object) is
   begin
      Obj.Curr := Obj.Head;
   end Reset;

   -- Reset the iterator
   procedure Reset (Ptr: in out not null Pointer) is
   begin
      Reset(Ptr.all);
   end Reset;

   -- Query if the iterator has finished
   function Has_Next (Obj: in Object) return Boolean is
   begin
      return (Obj.Curr /= null);
   end Has_Next;

   -- Query if the iterator has finished
   function Has_Next (Ptr: in not null Pointer) return Boolean is
   begin
      return Has_Next(Ptr.all);
   end Has_Next;

   -- Return the next item from the list
   procedure Get_Next (Obj: in out Object; Next_Item: out Item) is
      Act: Container_Ptr := Obj.Curr;
   begin
      if (Act /= null) then
         Obj.Curr := Obj.Curr.Succ;
         Next_Item := Act.Data;
      else
         raise List_Exception;
      end if;
   end Get_Next;

   -- Return the next item from the list
   procedure Get_Next (Ptr: in out not null Pointer; Next_Item: out Item) is
   begin
      Get_Next(Ptr.all, Next_Item);
   end Get_Next;

   -- Query the element count
   function Count (Obj: in Object) return Long_Long_Natural is
   begin
      return Obj.Count;
   end Count;

   -- Query the element count
   function Count (Ptr: in not null Pointer) return Long_Long_Natural is
   begin
      return Count(Ptr.all);
   end Count;

   procedure Push (Obj: in out Object; I: in Item) is
   begin
      Append(Obj, I);
   end Push;

   procedure Push (Ptr: in out not null Pointer; I: in Item) is
   begin
      Push(Ptr.all, I);
   end Push;

   procedure Pop (Obj: in out  Object; I: out Item) is
      Aux: Container_Ptr;
   begin
      if not Empty(Obj) then
         I := Obj.Tail.Data;
         Aux := Obj.Tail.Prev;
         if Aux /= null then
            Aux.Succ := null;
         end if;
         Obj.Count := Obj.Count - 1;
         Free_Container(Obj.Tail);
         Obj.Tail := Aux;
         if (Obj.Tail = null) then
            Init(Obj);
         end if;
      end if;
   end Pop;

   procedure Pop (Ptr: in out not null Pointer; I: out Item) is
   begin
      Pop(Ptr.all, I);
   end Pop;

   procedure Peek (Obj: in out Object; I: out Item) is
   begin
      if not Empty(Obj) then
         I := Obj.Tail.Data;
      end if;
   end Peek;

   procedure Peek (Ptr: in out not null Pointer; I: out Item) is
   begin
      Peek(Ptr.all, I);
   end Peek;

   function Empty (Obj: in Object) return Boolean is
   begin
      return (Count(Obj) = 0);
   end Empty;

   function Empty (Ptr: in not null Pointer) return Boolean is
   begin
      return Empty(Ptr.all);
   end Empty;

end Ast2Cfg.Generic_List;
