---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--          Copyright (c) 2007, Raul Fechete, Georg Kienesberger         --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

generic

   type Item is private;

package Ast2Cfg.Generic_List is

   type Object is private;
   type Pointer is access Object; -- only use if you _really_ need a pointer

   List_Exception: exception;

   -- List elements methods
   procedure Init         (Obj:    out Object);
   procedure Init         (Ptr:    out Pointer);
   procedure Append       (Obj: in out Object; I: in Item);
   procedure Append       (Obj: in out Object; O: in out Object);
   procedure Append       (Ptr: in out not null Pointer; I: in Item);
   procedure Append       (Ptr: in out not null Pointer; P: in out not null Pointer);
   procedure Insert_Head  (Obj: in out Object; I: in Item);
   procedure Insert_Head  (Ptr: in out not null Pointer; I: in Item);
   function  Get_Head     (Obj: in     Object) return Item;
   function  Get_Head     (Ptr: in     not null Pointer) return Item;
   function  Get_Tail     (Obj: in     Object) return Item;
   function  Get_Tail     (Ptr: in     not null Pointer) return Item;
   procedure Replace_Head (Obj: in out Object; I: in Item);
   procedure Replace_Head (Ptr: in out not null Pointer; I: in Item);
   procedure Remove_Last  (Obj: in out Object);
   procedure Remove_Last  (Ptr: in out not null Pointer);
   procedure Remove       (Obj: in out Object; I: in Item); -- removes the first element equal to I
   procedure Remove       (Ptr: in out not null Pointer; I: in Item); -- removes the first element equal to I
   procedure Duplicate    (Src: in out Object; Dst: out Object);
   procedure Duplicate    (Src: in out not null Pointer; Dst: out Pointer);
   procedure Clear        (Obj: in out Object); -- clears the list
   procedure Clear        (Ptr: in out not null Pointer); -- clears the list
   procedure Free         (Ptr: in out Pointer); -- clears the list and frees the
                                                 -- list header (Object)

   -- Stack-like methods
   procedure Push      (Obj: in out Object; I: in Item);
   procedure Push      (Ptr: in out not null Pointer; I: in Item);
   procedure Pop       (Obj: in out Object; I: out Item);
   procedure Pop       (Ptr: in out not null Pointer; I: out Item);
   procedure Peek      (Obj: in out Object; I: out Item);
   procedure Peek      (Ptr: in out not null Pointer; I: out Item);
   function  Empty     (Obj: in Object) return Boolean;
   function  Empty     (Ptr: in not null Pointer) return Boolean;

   -- Iterator methods
   procedure Reset (Obj: in out Object);
   procedure Reset (Ptr: in out not null Pointer);
   function  Has_Next (Obj: in Object) return Boolean;
   function  Has_Next (Ptr: in not null Pointer) return Boolean;
   procedure Get_Next (Obj: in out Object; Next_Item: out Item);
   procedure Get_Next (Ptr: in out not null Pointer; Next_Item: out Item);
   function  Count (Obj: in Object) return Long_Long_Natural;
   function  Count (Ptr: in not null Pointer) return Long_Long_Natural;

private

   type Item_Ptr is access Item;

   type Container;
   type Container_Ptr is access Container;
   type Container is
      record
         Data:      Item;
         Prev,Succ: Container_Ptr := null;
      end record;

   type Object is
      record
         Head:  Container_Ptr := null;
         Tail:  Container_Ptr := null;
         Curr:  Container_Ptr := null;
         Count: Long_Long_Natural := 0;
      end record;

end Ast2Cfg.Generic_List;
